/* shadowfs is Copyright (C) 2003 - 2004 Clifford Wolf
 * shadowfs is Copyright (C) 2003 - 2004 LINBIT Information Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. A copy of the GNU General
 * Public License can be found at COPYING.
 */

#include "config.h"

#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

void *get_dl_symbol(char * symname)
{
        void *rc;
#if DLOPEN_LIBC
        static void *libc_handle = 0;

        if (!libc_handle) libc_handle=dlopen("libc.so.6", RTLD_LAZY);
        if (!libc_handle) {
                printf("shadowfs: Can't dlopen libc: %s\n", dlerror());
                abort();
        }

        rc = dlsym(libc_handle, symname);
#  if DEBUG == 1
        fprintf(stderr, "shadowfs debug [%d]: Symbol '%s' in libc (%p) has been resolved to %p.\n",
                getpid(), symname, libc_handle, rc);
#  endif
        dlclose(libc_handle);
#else
        rc = dlsym(RTLD_NEXT, symname);
#  if DEBUG == 1
        fprintf(stderr, "shadowfs debug [%d]: Symbol '%s' (RTLD_NEXT) has been resolved to %p.\n",
                getpid(), symname, rc);
#  endif
#endif
        if (!rc) {
                fprintf(stderr, "shadowfs: Can't resolve %s: %s\n",
                       symname, dlerror());
                abort();
        }

#if DEBUG == 1
	{
		Dl_info dli;
		dladdr(__builtin_return_address(0), &dli);
		fprintf(stderr, "shadowfs debug [%d]: %s "
				"called by %p [ %s(%p) %s(%p) ].\n",
				getpid(), __func__,
				 __builtin_return_address(0),
				strrchr(dli.dli_fname, '/') ?
					strrchr(dli.dli_fname, '/')+1 : dli.dli_fname,
				dli.dli_fbase, dli.dli_sname, dli.dli_saddr); 
		dladdr(__builtin_return_address(1), &dli);
		fprintf(stderr, "shadowfs debug [%d]: %*s "
				"called by %p [ %s(%p) %s(%p) ].\n",
				getpid(), (int)strlen(__func__), "...",
				__builtin_return_address(1),
				strrchr(dli.dli_fname, '/') ?
					strrchr(dli.dli_fname, '/')+1 : dli.dli_fname,
				dli.dli_fbase, dli.dli_sname, dli.dli_saddr); 
	}
#endif

	return rc;
}

