/* shadowfs is Copyright (C) 2003 - 2004 Clifford Wolf
 * shadowfs is Copyright (C) 2003 - 2004 LINBIT Information Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. A copy of the GNU General
 * Public License can be found at COPYING.
 */

#include "config.h"
#include "getrealfn.h"

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <utime.h>

#include "wrapper_func.h"


/*
 * glibc's realpath() is broken - so don't use it!
 *
 * FIXME: This is not re-entrant! Let the caller free the return value..
 *
 * FIXME: Add a small cache to avoid constant lookups of the same path.
 * Don't forget to add a smart mechanism to flush the cache (e.g. on sleep
 * or wait system calls).
 */
const char *getrealfn(const char *filename)
{
	static char *ret = 0;
	char *srcfile, *token, *tb, *tp;
	char buffer[4096 + 1], *newfile;
	int recursion = 16;
	int len, gotlink;

	if ( ret ) free(ret);
	ret = strdup("");

	/* get absolute path name */
	if ( *filename == '/' ) srcfile = strdup(filename);
	else {
		if ( !getcwd(buffer, 4096) ) return filename;
		asprintf(&srcfile, "%s/%s", buffer, filename);
	}

#ifdef WRAPPED_READLINK
	resolve_readlink();
#endif

tail_recursion:
	gotlink = 0; tp = srcfile;
	while ( (token = strtok_r(tp, "/", &tb)) )
	{
		tp = 0;
		if ( *token == 0 ) continue;

		asprintf(&newfile, "%s/%s", ret, token);
#ifdef WRAPPED_READLINK
		len = orig_readlink(newfile, buffer, 4096);
#else
		len = readlink(newfile, buffer, 4096);
#endif

		if ( len >= 0 ) {
			buffer[len] = 0;
			free(newfile);
			if ( *buffer == '/' ) {
				free(ret);
				ret = strdup(buffer);
			} else {
				asprintf(&newfile, "%s/%s", ret, buffer);
				free(ret); ret = newfile;
			}
			gotlink = 1;
		} else {
			free(ret);
			ret = newfile;
		}
	}

	free(srcfile);

	if ( gotlink && recursion-- > 0 ) {
		srcfile = ret;
		ret = strdup("");
		goto tail_recursion;
	}

	tp = srcfile = ret; ret = strdup("");
	while ( (token = strtok_r(tp, "/", &tb)) ) {
		tp = 0;

		if ( !strcmp(token, "..") ) {
			newfile = strrchr(ret, '/');
			if ( newfile ) *newfile = 0;
		} else {
			asprintf(&newfile, "%s/%s", ret, token);
			free(ret); ret=newfile;
		}
	}

	if ( *ret == 0 ) {
		free(ret);
		ret = strdup("/");
	}

	return ret;
}

