/* shadowfs is Copyright (C) 2003 - 2004 Clifford Wolf
 * shadowfs is Copyright (C) 2003 - 2004 LINBIT Information Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. A copy of the GNU General
 * Public License can be found at COPYING.
 */

#include "config.h"
#include "wrapper.h"

#include <stdarg.h>

extern void *get_dl_symbol(char *);

struct utimbuf;
struct timeval;
struct stat;
typedef struct FILE FILE;

typedef int mode_t;
typedef int dev_t;
typedef int pid_t;

typedef __SIZE_TYPE__ size_t;

#if DEBUG_386 == 1

/*
 * This doesn't resolve any symbols and doesn't use
 * any libc features. So it's the perfect debug function
 * for our purposes.
 */
static void wrapper_debug_out(char *txt) {
	long len = 0;

	while (txt[len]) len++;

	asm volatile (
			"pushl %%ebx\n\t"
			"mov $2, %%ebx\n\t"	/* stderr */
			"int $0x80\n\t"
			"popl %%ebx"
		: /* output */
		: /* input */
			"a" ((long)(4)),	/* write  */
		//	"b" ((long)(2)),	/* stderr */
			"c" ((long)(txt)),	/* text   */
			"d" ((long)(len))	/* length */
	);
}

#else

static inline void wrapper_debug_out(char *txt) { return; }

#endif

#undef WRAPPER

#define WRAPPER(ret, name, args, xargs, action)		\
extern ret name args; 					\
ret (*orig_ ## name) args = 0;				\
void resolve_ ## name() {				\
	if ( ! orig_ ## name )				\
		orig_ ## name = get_dl_symbol(#name);	\
}							\
ret name args {						\
	char *F = #name;				\
	int P = 0, E = 0;				\
	ret R = 0;					\
	wrapper_debug_out("[[ " #name " ]]\n");		\
	resolve_ ## name();				\
	action;						\
	R = orig_ ## name xargs;			\
	P = 1;						\
	action;						\
	return R;					\
}

#define WRAPPER_OPEN(dummy, name, action)		\
int (*orig_ ## name)(const char *p, int f, ...) = 0;	\
void resolve_ ## name() {				\
	if ( ! orig_ ## name )				\
		orig_ ## name = get_dl_symbol(#name);	\
}							\
int name(const char *p, int f, ...) {			\
	char *F = #name;				\
	int P = 0, E = 0, R = 0;			\
	wrapper_debug_out("[[ " #name " ]]\n");		\
	resolve_ ## name();				\
	action;						\
	if (f & MY_CREAT) {				\
		va_list ap;				\
		mode_t m = 0;				\
							\
		va_start(ap, f);			\
		m = va_arg(ap, mode_t);			\
		va_end(ap);				\
							\
		R = orig_ ## name(p, f, m);		\
	} else						\
		R = orig_ ## name(p, f);		\
	P = 1;						\
	action;						\
	return R;					\
}

#include "wrapper_func.h"

